////////////////////////////////////////////////////////////////////////
// Class:       SignalNoiseAna
// Plugin Type: analyzer (art v3_05_01)
// File:        SignalNoiseAna_module.cc
//
// Generated at Sat Sep 19 08:41:53 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art/Utilities/make_tool.h"

#include "larcore/Geometry/Geometry.h"
#include "larcoreobj/SimpleTypesAndConstants/RawTypes.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"

#include "TEfficiency.h"
#include "TFile.h"
#include "TTree.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TCanvas.h"
#include "TProfile.h"
#include "TRandom.h"

using namespace std;

class SignalNoiseAna;


class SignalNoiseAna : public art::EDAnalyzer {
public:
  explicit SignalNoiseAna(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  SignalNoiseAna(SignalNoiseAna const&) = delete;
  SignalNoiseAna(SignalNoiseAna&&) = delete;
  SignalNoiseAna& operator=(SignalNoiseAna const&) = delete;
  SignalNoiseAna& operator=(SignalNoiseAna&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;

private:

  // User's function
  void beginJob() override;
  void endJob() override;
  void calculateFFT(TH1D* h1_waveform, TH1D* h1_fft);
  void getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick);
  bool isSignalInWindow(int maxtick, int starttick, int endtick);
  void reset();

  // fhicl parameters
  std::string fRawDigitLabel;
  art::InputTag fWireProducerLabel;
  art::InputTag fSimulationProducerLabel;
  bool fUseFullWaveform;
  int fShortWaveformSize;
  bool fUseNumElectrons; // true/flase: signal according to "number of electrons"/"energy deposits"
  double fSignalNumElectronsLowCut;
  double fSignalNumElectronsHighCut;
  bool fSaveSignal;
  bool fSaveNoise;
  bool fEvaluateWaveform;
  bool fDropLastWindow;
  int fStrideSize;
  std::string fFrequencyTemplateFile;
  bool fSaveTree;

  
  // user's parameters
  int fNPlanes;
  int fNTicks;
  float fFFTBinWidth;

  TH1D* h1_signal[3];

  TH2D* h2_signalADC_sametick[3];
  TH2D* h2_signalADC_max[3];
  TH1D* h1_tickdiff_max[3];
  
  TH1D* h1_noiserms[3];
  TProfile* pf_frequency_noise[3];
  TProfile* pf_frequency_signal[3];
  TProfile* pf_frequency_both[3];

  TProfile* pf_frequency_noise_default[3];
  TProfile* pf_frequency_signal_default[3];
  TProfile* pf_frequency_both_default[3];

  TProfile* pf_frequency_noise_template[3];
  TProfile* pf_frequency_signal_template[3];
  TProfile* pf_frequency_both_template[3];
  
  TH1D* h1_chisquare_noise[3];
  TH1D* h1_chisquare_signal[3];
  TH2D* h2_chisquare[3];

  TH1D* h1_test_waveform;
  TH1D* h1_test_fft;

  // tree
  TTree* fTree;
  int run;
  int subrun;
  int event;
  
  // waveform
  short channeladc[480][2048]; 
  double waveformsignal[480][14];
  double waveformchisquare[480][14];
  double frequencybinvalue[480][14][100];
  
  // signal
  std::vector<int> simchannelvec; // sim channel channel number per event
  std::vector<std::vector<int>> signalstarttickvec; // signals per each sim channel per event
  std::vector<std::vector<int>> signalendtickvec; // signals per each sim channel per event
  std::vector<std::vector<int>> signalmaxtickvec; // signals per each sim channel per event
  std::vector<std::vector<double>> signalmaxvec; // signals per each sim channel per event

};


SignalNoiseAna::SignalNoiseAna(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  fRawDigitLabel             = p.get<std::string>("RawDigitLabel", "");
  fWireProducerLabel         = p.get<art::InputTag>("WireProducerLabel", "");
  fSimulationProducerLabel   = p.get<art::InputTag>("SimulationProducerLabel","largeant");
  fUseFullWaveform           = p.get<bool>("UseFullWaveform", true);
  fShortWaveformSize         = p.get<int>("ShortWaveformSize", 200);
  fUseNumElectrons           = p.get<bool>("UseNumElectrons", true);
  fSignalNumElectronsLowCut  = p.get<double>("SignalNumElectronsLowCut");
  fSignalNumElectronsHighCut  = p.get<double>("SignalNumElectronsHighCut");
  fSaveSignal                = p.get<bool>("SaveSignal", false);
  fSaveNoise                 = p.get<bool>("SaveNoise", false);
  fEvaluateWaveform          = p.get<bool>("EvaluateWaveform", false);
  fDropLastWindow            = p.get<bool>("DropLastWindow", true);
  fStrideSize                = p.get<int>("StrideSize", 150);
  fFrequencyTemplateFile     = p.get<std::string>("FrequencyTemplateFile");
  fSaveTree                  = p.get<bool>("SaveTree", false);

  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("SignalNoiseAna") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("SignalNoiseAna") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }

  fNPlanes = lar::providerFrom<geo::Geometry>()->Nplanes();
  if (fNPlanes > 3) {
    throw cet::exception("SignalNoiseAna") << "Number of planes is " << fNPlanes << "; please check";
  }
  std::cout << "fNPlanes: " << fNPlanes << std::endl;

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);
  fNTicks = detProp.NumberTimeSamples(); // number of clock ticks per event
  float fSampleRate = sampling_rate(clockData); // period of the TPC readout electronics clock [us]; 1 tick = 1 tdc = 198 ns.
  cout << "Numer of clock ticks per event: " << fNTicks << endl;
  cout << "Sampling rate: " << fSampleRate << endl;
  // width of frequency bin in kHz (signal window is 2048 here)
  fFFTBinWidth = 1.0/(fNTicks*fSampleRate*1.0e-6); // [kHz]
  cout << "fFFTBinWidth: " << fFFTBinWidth << endl;
  
}

void SignalNoiseAna::analyze(art::Event const& e)
{
  if (fSaveTree) reset();

  run = e.run();
  subrun = e.subRun();
  event = e.id().event();

  auto const* geo = lar::providerFrom<geo::Geometry>();
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataFor(e);
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataFor(e, clockData);
  auto const& chStatus = art::ServiceHandle<lariov::ChannelStatusService>()->GetProvider();
  
  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }

  // get wire
  art::Handle<std::vector<recob::Wire>> wireListHandle;
  std::vector<art::Ptr<recob::Wire>> wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  auto simChannelHandle = e.getValidHandle<std::vector<sim::SimChannel>>(fSimulationProducerLabel);

  if (fSaveSignal) {
    // ... Loop over simChannels
    for (auto const& simchannel : (*simChannelHandle)) {

      // .. get simChannel channel number
      const raw::ChannelID_t channel = simchannel.Channel();
      if (chStatus.IsBad(channel)) 
        continue;

      int view = geo->View(channel);

      std::vector<int> signal_starttick;
      std::vector<int> signal_endtick;
      std::vector<double> signal_max; // according to number of electrons or energy deposits
      std::vector<int> signal_max_tdctick;

      getSignal(simchannel, signal_starttick, signal_endtick, signal_max, signal_max_tdctick);

      if (signal_starttick.size() == 0 ||
          (signal_endtick.size() == 1 && signal_endtick.back() == -999))
        continue;
      
      // ---- signal ADC ----
      std::vector<short> adcvec(fNTicks);
      // recob wire
      if (wirelist.size()) {
        for (size_t w = 0; w < wirelist.size(); w++) {
          if (wirelist[w]->Channel() != channel) continue;

          std::vector<float> vecADC =
            wirelist[w]->Signal(); // a zero-padded full length vector filled with RoI signal.
          
          for (size_t j = 0; j < vecADC.size(); ++j) {
            adcvec[j] = vecADC[j];
          }
        } // loop over wirelist w
      } // if recob wire
      else if (rawdigitlist.size()) {
        for (size_t r = 0; r < rawdigitlist.size(); r++) {
          if (rawdigitlist[r]->Channel() != channel) continue;

          raw::RawDigit::ADCvector_t rawadc(fNTicks);
          raw::Uncompress(rawdigitlist[r]->ADCs(), rawadc, rawdigitlist[r]->Compression());

          float pedestal = rawdigitlist[r]->GetPedestal();
          for (size_t j=0; j < rawadc.size(); ++j) {
            adcvec[j] = rawadc[j] - pedestal;
          }
        } // loop over rawdigitlist r
      } // if rawdigit
      
      // "max" signal
      std::vector<short>::iterator temp_index = std::max_element(adcvec.begin(), adcvec.end());
      float max_adc = *temp_index;
      //if (max_adc == 0) continue; // skip those sim channel with "poor" simulation
      int max_adc_tick = std::distance(adcvec.begin(), temp_index);

      int max_signal_index = std::distance(signal_max.begin(), std::max_element(signal_max.begin(), signal_max.end()));
      
      h1_signal[view]->Fill(signal_max[max_signal_index]);
      h2_signalADC_sametick[view]->Fill(signal_max[max_signal_index], adcvec[signal_max_tdctick[max_signal_index]]);
      h2_signalADC_max[view]->Fill(signal_max[max_signal_index], max_adc);
      h1_tickdiff_max[view]->Fill(signal_max_tdctick[max_signal_index] - max_adc_tick);

      // signal FFT
      if (signal_max[max_signal_index] < fSignalNumElectronsLowCut) continue;
      
      if (fUseFullWaveform) {
        if (fSignalNumElectronsHighCut != -1 && signal_max[max_signal_index] > fSignalNumElectronsHighCut) continue;

        TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fNTicks, 0, fNTicks);
        for (int j = 0; j < fNTicks; ++j) {
          h1_waveform->SetBinContent(j+1, adcvec[j]);
        }
        TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fNTicks, 0, fNTicks*fFFTBinWidth);

        // do FFT for single waveform
        calculateFFT(h1_waveform, h1_fft); 

        // save for each plane
        for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
          pf_frequency_signal[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          pf_frequency_signal_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
        }

        delete h1_waveform;
        delete h1_fft;
      } // fUseFullWaveform
      else {
        // loop over signals
        int tdc_start = -1;
        int tdc_end = -1;
        for (size_t s = 0; s < signal_starttick.size(); ++s) {

          if (signal_max[s] < fSignalNumElectronsLowCut) continue;
          if (fSignalNumElectronsHighCut != -1 && signal_max[s] > fSignalNumElectronsHighCut) continue;

          int start_s = s;
          int last_s = s;

          // check the following signals
          for (size_t s2 = s + 1; s2 < signal_starttick.size(); ++s2) {
            
            if (signal_max[s2] < fSignalNumElectronsLowCut) continue;
            if (fSignalNumElectronsHighCut != -1 && signal_max[s2] > fSignalNumElectronsHighCut) continue;

            if (signal_max_tdctick[s2] >= signal_max_tdctick[s] + fShortWaveformSize) {
              s = s2 - 1;
              break;
            }
            else {
              last_s = s2;
              if (s2 == signal_starttick.size() - 1) s = s2; 
            }
          } // loop over signal s2
          
          // select a random range based on dt
          int dt = signal_max_tdctick[last_s] - signal_max_tdctick[start_s];
          
          tdc_start = signal_max_tdctick[start_s] - ((fShortWaveformSize - dt - 0.1 * fShortWaveformSize) / 2) * gRandom->Uniform(0,1);

          if (tdc_start < tdc_end) tdc_start = tdc_end + 1; // do not overlap with previous one
          if (tdc_start < 0) tdc_start = 0;

          tdc_end = tdc_start + fShortWaveformSize - 1;
          if (tdc_end > fNTicks - 1) {
            tdc_end = fNTicks -1;
            tdc_start = tdc_end - fShortWaveformSize + 1;
            // ideally, another check with previous one is needed here, but this should be very rare
          }

          // short waveform and FFT
          TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fShortWaveformSize, 0, fShortWaveformSize);
          for (int j = tdc_start; j <= tdc_end; ++j) {
            h1_waveform->SetBinContent(j-tdc_start+1, adcvec[j]);
          }
          TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
        
          // do FFT for each short waveform
          calculateFFT(h1_waveform, h1_fft); 

          // save for each plane
          for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
            pf_frequency_signal[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
            pf_frequency_signal_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          }

          delete h1_waveform;
          delete h1_fft;

        } // loop over signal s
      } // use short waveform

    } // loop over simChannels
  } // fSaveSignal
  
  // ----------- noise -----------------
  if (fSaveNoise) {
    std::map<raw::ChannelID_t, bool> signalMap;
    for (auto const& channel : (*simChannelHandle)) {
      signalMap[channel.Channel()] = true;
    }

    for (size_t i = 0; i < (rawdigitlist.empty() ? wirelist.size() : rawdigitlist.size()); ++i) {
      std::vector<short> adcvec(fNTicks);

      int channel = -1;

      if (rawdigitlist.size()) {
        channel = rawdigitlist[i]->Channel();
        if (signalMap[channel]) continue;
        if (chStatus.IsBad(channel)) continue;

        std::vector<short> rawadc(fNTicks); // vector to hold uncompressed adc values later
        raw::Uncompress(rawdigitlist[i]->ADCs(), rawadc, rawdigitlist[i]->GetPedestal(), rawdigitlist[i]->Compression());
        for (size_t j = 0; j < rawadc.size(); ++j) {
          adcvec[j] = rawadc[j] - rawdigitlist[i]->GetPedestal();
        }
      }
      else if (wirelist.size()) {
        channel = wirelist[i]->Channel();
        if (signalMap[channel]) continue;
        if (chStatus.IsBad(channel)) continue;

        const auto& signal = wirelist[i]->Signal();
        for (size_t j = 0; j < adcvec.size(); ++j) {
          adcvec[j] = signal[j];
        }
      }

      int view = geo->View(channel);

      // noise rms
      double temp_sum = 0.;
      for (size_t j = 0; j < adcvec.size(); j++) {
        temp_sum +=  adcvec[j] * adcvec[j];
      }
      h1_noiserms[view]->Fill(sqrt(temp_sum / adcvec.size()));

      // waveform and frequency
      if (fUseFullWaveform) {
        TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fNTicks, 0, fNTicks);
        for (int j = 0; j < fNTicks; ++j) {
          h1_waveform->SetBinContent(j+1, adcvec[j]);
        }
        TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fNTicks, 0, fNTicks*fFFTBinWidth);

        // do FFT for single waveform
        calculateFFT(h1_waveform, h1_fft); 

        // save for each plane
        for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
          pf_frequency_noise[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          pf_frequency_noise_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
        }

        delete h1_waveform;
        delete h1_fft;
      } // if fUseFullWaveform
      else {
        // only choose one fShortWaveformSize waveform
        //int start_tick = int((fNTicks - fShortWaveformSize) * gRandom->Uniform(0, 1)); 
        // save more waveform: we do not save all possible short waveform, but choose a subset randomly 
        int nShortWaveformSaved = int (fNTicks / fShortWaveformSize) - 1;
        int start_tick = int((fNTicks % fShortWaveformSize + fShortWaveformSize) * gRandom->Uniform(0, 1)); 
        for (int k = 0; k < nShortWaveformSaved; k++) {
          TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fShortWaveformSize, 0, fShortWaveformSize);
          for (int j = (start_tick + k * fShortWaveformSize); j < (start_tick + (k + 1) * fShortWaveformSize); ++j) {
            h1_waveform->SetBinContent(j - start_tick - k * fShortWaveformSize + 1, adcvec[j]);
          }
          TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
        
          // do FFT for each short waveform
          calculateFFT(h1_waveform, h1_fft); 

          // save for each plane
          for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
            pf_frequency_noise[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
            pf_frequency_noise_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          }

          delete h1_waveform;
          delete h1_fft;
        } // loop over k
      } // if not fUseFullWaveform


    } // loop over rawgigit or wire
  } // fSaveNoise
  
  // ---------- evaluate a waveform ----------
  if (fEvaluateWaveform) {
    for (size_t i = 0; i < (rawdigitlist.empty() ? wirelist.size() : rawdigitlist.size()); ++i) {

      // test waveform
      std::vector<short> adcvec(fNTicks);
      int channel = -1;

      if (rawdigitlist.size()) {
        channel = rawdigitlist[i]->Channel();
        if (chStatus.IsBad(channel)) continue;

        std::vector<short> rawadc(fNTicks);
        raw::Uncompress(rawdigitlist[i]->ADCs(), rawadc, rawdigitlist[i]->GetPedestal(), rawdigitlist[i]->Compression());
        for (size_t j = 0; j < rawadc.size(); ++j) {
          adcvec[j] = rawadc[j] - rawdigitlist[i]->GetPedestal();
          if (fSaveTree) channeladc[channel][j] = adcvec[j];
        }
      }
      else if (wirelist.size()) {
        channel = wirelist[i]->Channel();
        if (chStatus.IsBad(channel)) continue;

        const auto& signal = wirelist[i]->Signal();
        for (size_t j = 0; j < adcvec.size(); ++j) {
          adcvec[j] = signal[j];
          if (fSaveTree) channeladc[channel][j] = adcvec[j];
        }
      }

      int view = geo->View(channel);

      bool IsSignalWire = false;
      std::vector<int> signal_starttick;
      std::vector<int> signal_endtick;
      std::vector<double> signal_max; // according to number of electrons or energy deposits
      std::vector<int> signal_max_tdctick;
      for (auto const& simchannel : (*simChannelHandle)) {
        // .. get simChannel channel number
        if (int(simchannel.Channel()) != channel) continue;
        
        getSignal(simchannel, signal_starttick, signal_endtick, signal_max, signal_max_tdctick);

        if (signal_starttick.size() == 0 ||
            (signal_endtick.size() == 1 && signal_endtick.back() == -999))
          continue;

        IsSignalWire = true;
        
        if (fSaveTree) {
          simchannelvec.push_back(channel);
          signalstarttickvec.push_back(signal_starttick);
          signalendtickvec.push_back(signal_endtick);
          signalmaxtickvec.push_back(signal_max_tdctick);
          signalmaxvec.push_back(signal_max);
        }

      } // loop over simchannel

      
      if (fUseFullWaveform) {
        cout << "www:  EvaluateWaveform using full waveform needs to be added!" << endl;
      } // if fUseFullWaveform
      else {
        int nStrides = int(fNTicks / fStrideSize + 1);
        if (fDropLastWindow) nStrides = nStrides - 1;
        for (int k = 0; k < nStrides; ++k) {
          bool IsSignalWindow = false;
          
          int start_tick = fStrideSize * k;
          int end_tick = start_tick + fShortWaveformSize - 1;
          if (end_tick > fNTicks - 1) {
            end_tick = fNTicks - 1;
            start_tick = end_tick + 1 - fShortWaveformSize;
          }

          TH1D* h1_waveform = new TH1D(Form("waveform_%d_%d", channel, k), Form("waveform_%d_%d", channel, k), fShortWaveformSize, 0, fShortWaveformSize);
          for (int j = start_tick; j <= end_tick; ++j) {
            h1_waveform->SetBinContent(j-start_tick+1, adcvec[j]);
          }
          
          TH1D* h1_fft = new TH1D(Form("fft_%d_%d", channel, k), Form("fft_%d_%d", channel, k), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);

          calculateFFT(h1_waveform, h1_fft);
          
          // save test
          if (e.id().event() == 1 && channel == 100 &&  k==6) {
            for (int j = 0; j < fShortWaveformSize; ++j) {
              h1_test_waveform->SetBinContent(j+1, h1_waveform->GetBinContent(j+1));
              h1_test_fft->SetBinContent(j+1, h1_fft->GetBinContent(j+1));
            }
          }
          
          // chi-square according to template
          double chi2 = 0.;
          for (int j = 0; j < fShortWaveformSize/2; ++j) {
            double value = (pf_frequency_noise_template[view]->GetBinContent(j+1) - h1_fft->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
            chi2 += value*value;
            frequencybinvalue[channel][k][j] = value*value;
          }
          //cout << "k: " << k<< "; chi2: " << chi2 << "; chi2 / NDF: " << chi2 / (fShortWaveformSize/2) << endl;
         
          double chi2ndf = chi2 / (fShortWaveformSize/2);
          if (chi2ndf > 10.) chi2ndf = 10. - 1e-6;

          double signal = 0.;
          if (IsSignalWire) {
            for (size_t s = 0; s < signal_starttick.size(); ++s) {
              if(isSignalInWindow(signal_max_tdctick[s], start_tick, end_tick)) {
                if (signal_max[s] > signal) {
                  signal = signal_max[s];
                }
              }
            }
          }
          
          waveformsignal[channel][k] = signal;
          waveformchisquare[channel][k] = chi2ndf;

          if (signal > 0.) {
            IsSignalWindow = true;
            if (signal < 1000.) IsSignalWindow = false;
          }

          if (IsSignalWindow) {
            h1_chisquare_signal[view]->Fill(chi2ndf);
            h2_chisquare[view]->Fill(1., chi2ndf);
          }
          // need include those ones on signal wire
          //else {
          if (!IsSignalWire) {
            h1_chisquare_noise[view]->Fill(chi2ndf);
            h2_chisquare[view]->Fill(0., chi2ndf);
          }

          delete h1_waveform;
          delete h1_fft;

        } // loop over nStrides k

      } // use short waveform
    } // loop over rawdigitlist or wirelist
  
    if (fSaveTree) fTree->Fill();
    
  } // if fEvaluateWaveform
}

void SignalNoiseAna::beginJob()
{
  art::ServiceHandle<art::TFileService const> tfs;

  TFile TemplateFFTFile(fFrequencyTemplateFile.c_str());
  
  for (int i = 0; i < fNPlanes; ++i) {
    if (fUseNumElectrons) {
      h1_signal[i] = tfs->make<TH1D>(Form("h1_signal_%d", i), Form("Plane %d; Number of Electrons;", i), 100, 0, 20000);
      h2_signalADC_sametick[i] = tfs->make<TH2D>(Form("h2_signalADC_sametick_%d",i), Form("Plane %d; number of electrons; ADC counts", i), 100, 0, 20000, 50, 0, 50);
      h2_signalADC_max[i] = tfs->make<TH2D>(Form("h2_signalADC_max_%d",i), Form("Plane %d; number of electrons; ADC counts", i), 100, 0, 20000, 50, 0, 50);
    }
    else {
      h1_signal[i] = tfs->make<TH1D>(Form("h1_signal_%d", i), Form("Plane %d; Energy [MeV];", i), 100, 0, 1);
      h2_signalADC_sametick[i] = tfs->make<TH2D>(Form("h2_signalADC_sametick_%d",i), Form("Plane %d; Energy [MeV]; ADC counts", i), 100, 0, 1, 50, 0, 50);
      h2_signalADC_max[i] = tfs->make<TH2D>(Form("h2_signalADC_max_%d",i), Form("Plane %d; Energy [MeV]; ADC counts", i), 100, 0, 1, 50, 0, 50);
    }
  
    h1_tickdiff_max[i] = tfs->make<TH1D>(Form("h1_tickdiff_max_%d", i), Form("Plane %d; tick diff (maxE - max pulse);", i), 100, -50, 50);
  

    h1_noiserms[i] = tfs->make<TH1D>(Form("h1_noiserms_%d",i), Form("Plane %d; ADC counts;", i), 200, 0, 10);

    if (fUseFullWaveform) {
      // TProfile: option "s" uses RMS as error; defauls uses RMS/sqrt(N) as error
      pf_frequency_noise[i] = tfs->make<TProfile>(Form("pf_frequency_noise_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth, "s");
      pf_frequency_signal[i] = tfs->make<TProfile>(Form("pf_frequency_signal_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth, "s");
      pf_frequency_both[i] = tfs->make<TProfile>(Form("pf_frequency_both_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth, "s");

      pf_frequency_noise_default[i] = tfs->make<TProfile>(Form("pf_frequency_noise_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth);
      pf_frequency_signal_default[i] = tfs->make<TProfile>(Form("pf_frequency_signal_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth);
      pf_frequency_both_default[i] = tfs->make<TProfile>(Form("pf_frequency_both_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth);
    }
    else {
      pf_frequency_noise[i] = tfs->make<TProfile>(Form("pf_frequency_noise_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth, "s");
      pf_frequency_signal[i] = tfs->make<TProfile>(Form("pf_frequency_signal_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth, "s");
      pf_frequency_both[i] = tfs->make<TProfile>(Form("pf_frequency_both_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth, "s");

      pf_frequency_noise_default[i] = tfs->make<TProfile>(Form("pf_frequency_noise_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
      pf_frequency_signal_default[i] = tfs->make<TProfile>(Form("pf_frequency_signal_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
      pf_frequency_both_default[i] = tfs->make<TProfile>(Form("pf_frequency_both_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
    }

    if (fEvaluateWaveform) {
      pf_frequency_noise_template[i] = (TProfile*)TemplateFFTFile.Get(Form("signalnoiseana/pf_frequency_noise_%d",i));
      pf_frequency_signal_template[i] = (TProfile*)TemplateFFTFile.Get(Form("signalnoiseana/pf_frequency_signal_%d",i));
      pf_frequency_both_template[i] = (TProfile*)TemplateFFTFile.Get(Form("signalnoiseana/pf_frequency_both_%d",i));

      h1_chisquare_noise[i] = tfs->make<TH1D>(Form("h1_chisquare_noise_%d", i), Form("Plane %d; #chi^{2}/NDF; Count [#]", i), 100, 0, 10);
      h1_chisquare_signal[i] = tfs->make<TH1D>(Form("h1_chisquare_signal_%d", i), Form("Plane %d; #chi^{2}/NDF; Count [#]", i), 100, 0, 10);
      h2_chisquare[i] = tfs->make<TH2D>(Form("h2_chisquare_%d", i), Form("Plane %d; Signal/Noise Index; #chi^{2}/NDF", i), 5, 0, 5, 100, 0, 10);
    }
  }

  if (fEvaluateWaveform) {
    if (fUseFullWaveform) {
      h1_test_waveform = tfs->make<TH1D>("waveform_test", "waveform_test", fNTicks, 0, fNTicks);
      h1_test_fft = tfs->make<TH1D>("fft_test", "fft_test", fNTicks, 0, fNTicks*fFFTBinWidth);
    }
    else {
      h1_test_waveform = tfs->make<TH1D>("waveform_test", "waveform_test", fShortWaveformSize, 0, fShortWaveformSize);
      h1_test_fft = tfs->make<TH1D>("fft_test", "fft_test", fShortWaveformSize, 0, fShortWaveformSize*fFFTBinWidth);
    }
    
    if (fSaveTree) {
      fTree = tfs->make<TTree>("anatree", "signalnoiseana");
      fTree->Branch("run", &run, "run/I");
      fTree->Branch("subrun", &subrun, "subrun/I");
      fTree->Branch("event", &event, "event/I");

      fTree->Branch("channeladc", channeladc, "channeladc[480][2048]/S");
      fTree->Branch("waveformsignal", waveformsignal, "waveformsignal[480][14]/D");
      fTree->Branch("waveformchisquare", waveformchisquare, "waveformchisquare[480][14]/D");
      fTree->Branch("frequencybinvalue", frequencybinvalue, "frequencybinvalue[480][14][100]/D");

      fTree->Branch("simchannelvec", &simchannelvec);
      fTree->Branch("signalstarttickvec", &signalstarttickvec);
      fTree->Branch("signalendtickvec", &signalendtickvec);
      fTree->Branch("signalmaxtickvec", &signalmaxtickvec);
      fTree->Branch("signalmaxvec", &signalmaxvec);
    }
  }

}


void SignalNoiseAna::endJob()
{
  mf::LogInfo("SignalNoiseAna") << "endJob";
}


void SignalNoiseAna::calculateFFT(TH1D* h1_waveform, TH1D* h1_fft) {
  int nbins = h1_waveform->GetNbinsX();

  TH1* h1_transform = 0;

  h1_transform = h1_waveform->FFT(h1_transform, "MAG");

  // loop over h1_transform
  for (int k=0; k<nbins; k++) {
    double frequency = (k+0.5) * fFFTBinWidth; // [kHz]
    double amplitude = h1_transform->GetBinContent(k+1);
    h1_fft->Fill(frequency, amplitude);
  }

  h1_transform->Delete();
}


void SignalNoiseAna::getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick) {

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);

  auto const& timeSlices = simchannel.TDCIDEMap();

  // time slice from simChannel is for individual tick
  // group neighboring time slices into a "signal"
  int tdctick_previous = -999;
  double maxE = -999.; // number of electrons or energy deposits
  int maxEtick = -999;
  
  for (auto const& timeSlice : timeSlices) {
    auto const tpctime = timeSlice.first;
    auto const& energyDeposits = timeSlice.second;
    int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
    if (tdctick < 0 || tdctick > int(detProp.ReadOutWindowSize()) - 1) continue;

    // for a time slice, there may exist more than one energy deposit.
    double tdc_signal = 0;
    for (auto const& energyDeposit : energyDeposits) {
      if (fUseNumElectrons) {
        tdc_signal += energyDeposit.numElectrons;
      }
      else {
        tdc_signal += energyDeposit.energy;
      }
    }

    if (tdctick_previous == -999) {
      signal_starttick.push_back(tdctick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous != 1) {
      signal_starttick.push_back(tdctick);
      signal_endtick.push_back(tdctick_previous);
      signal_max.push_back(maxE);
      signal_max_tdctick.push_back(maxEtick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous == 1) {
      if (maxE < tdc_signal) {
        maxE = tdc_signal;
        maxEtick = tdctick;
      }
    }

    tdctick_previous = tdctick;

  } // loop over timeSlices timeSlice

  signal_endtick.push_back(tdctick_previous); // for last one
  signal_max.push_back(maxE);          // for last one
  signal_max_tdctick.push_back(maxEtick);     // for last one
}


bool SignalNoiseAna::isSignalInWindow(int maxtick, int starttick, int endtick) {
  // here, endtick included in the range
  return starttick <= maxtick && maxtick <= endtick;
}


void SignalNoiseAna::reset() {
  run = -99;
  subrun = -99;
  event = -99;
  
  for (int i = 0; i < 480; ++i) {
    for (int j = 0; j < 480; ++j) {
      channeladc[i][j] = -9999;
    }

    for (int k = 0; k < 14; ++k) {
      waveformsignal[i][k] = -9999;
      waveformchisquare[i][k] = -9999;
      for (int n=0; n<100; n++) {
        frequencybinvalue[i][k][n] = -9999;
      }
    }

  }

  simchannelvec.clear();
  signalstarttickvec.clear();
  signalendtickvec.clear();
  signalmaxtickvec.clear();
  signalmaxvec.clear();
}

DEFINE_ART_MODULE(SignalNoiseAna)
